# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

__version__ = '8.0.4'

__protocol_version__ = '2.1.0'
__control_protocol_version__ = '1.0.0'

# These are *protocol* versions for each package, *not* npm versions. To check, look at each package's src/version.ts file for the protocol version the package implements.
__jupyter_widgets_base_version__ = '2.0.0'
__jupyter_widgets_output_version__ = '1.0.0'
__jupyter_widgets_controls_version__ = '2.0.0'

# A compatible @jupyter-widgets/html-manager npm package semver range
__html_manager_version__ = '^1.0.1'
